import logging
from collections import defaultdict
from collections.abc import Iterator
from math import isclose
from pathlib import Path
from typing import Any, DefaultDict

import numpy as np
import pydicom
import SimpleITK

from panimg.exceptions import UnconsumedFilesException
from panimg.models import (
    EXTRA_METADATA,
    SimpleITKImage,
    validate_metadata_value,
)

logger = logging.getLogger(__name__)

OPTIONAL_METADATA_FIELDS = (
    # These fields will be included in the output mha file
    "Laterality",
    "SliceThickness",
    "WindowCenter",
    "WindowWidth",
    *[md.keyword for md in EXTRA_METADATA],
)


def format_error(message: str) -> str:
    return f"Dicom image builder: {message}"


class DicomTagNotFoundError(KeyError):
    pass


def _find_dicom_tag(dataset: pydicom.Dataset, tag: str):
    for header in dataset.iterall():
        if header.keyword == tag:
            return header.value

    raise DicomTagNotFoundError(
        f"Could not find DICOM tag {tag} in the header"
    )


class PixelValueInverter:
    """Inverts pixel values to generate MONOCHROME2 images from MONOCHROME1 images"""

    def __init__(self, image: np.ndarray):
        self.offset = np.add(np.min(image), np.max(image))
        self.dtype = image.dtype

    def invert(self, pixel_array: np.ndarray) -> np.ndarray:
        a = np.asarray(pixel_array, dtype=self.dtype)
        return np.add(-a, self.offset)  # use np.add to avoid overflow warnings


class DicomDataset:
    def __init__(self, *, name, headers, n_time, n_slices, n_slices_per_file):
        self.name = name
        self.headers = headers
        self.n_time = n_time
        self.n_slices = n_slices
        self.n_slices_per_file = n_slices_per_file

        self._direction = None
        self._pixel_value_dtype = None
        self._pixel_value_inverter = None

    @property
    def ref_header(self) -> pydicom.Dataset:
        return self.headers[0]["data"]

    @property
    def dimensions(self) -> int:
        # Images are either 4D or 3D (2D images get an additional axis of size 1)
        return 4 if self.n_time and self.n_time > 1 else 3

    @property
    def direction(self) -> np.ndarray:
        if self._direction is None:
            # Compute rotation matrix (orientation of the image)
            try:
                orientation = _find_dicom_tag(
                    self.ref_header, "ImageOrientationPatient"
                )
                row_cos = orientation[:3]
                col_cos = orientation[3:]
            except DicomTagNotFoundError:
                # Tag can be missing in X-ray images for example
                row_cos = (1, 0, 0)
                col_cos = (0, 1, 0)

            direction = np.eye(self.dimensions, dtype=float)
            direction[:3, :3] = np.stack(
                [row_cos, col_cos, np.cross(row_cos, col_cos)], axis=1
            )
            self._direction = direction

        return self._direction

    def _iter_origins(self):
        has_frame_details = (
            "PerFrameFunctionalGroupsSequence" in self.ref_header
        )
        if self.n_slices_per_file == 1 or not has_frame_details:
            # One or multiple regular DICOM files
            for partial in self.headers:
                if "ImagePositionPatient" in partial["data"]:
                    yield np.array(
                        partial["data"].ImagePositionPatient, dtype=float
                    )
        else:
            # An enhanced DICOM file with the entire volume stored in one file
            for frame in self.ref_header.PerFrameFunctionalGroupsSequence:
                try:
                    file_origin = _find_dicom_tag(
                        frame, "ImagePositionPatient"
                    )
                except DicomTagNotFoundError:
                    pass
                else:
                    yield np.array(file_origin, dtype=float)

    def _sort_slices_by_instance_number(self):
        """
        Sorts the slices of this study according to InstanceNumber.

        Raises
        ------
        ValueError
            If for any slice the InstanceNumber value is missing or
            not a number.
        """
        if len(self.headers) == 1:
            return  # no need to sort if there is only a single file

        try:
            self.headers.sort(key=lambda x: int(x["data"].InstanceNumber))
        except (TypeError, AttributeError) as e:
            # InstanceNumber is missing, empty or None but is needed to sort
            # the slices (could also sort by coordinates, but that is a lot
            # more complicated so currently not implemented)
            raise ValueError(
                "Could not determine slice order "
                "due to missing or invalid InstanceNumber"
            ) from e

    def _determine_slice_order(self):
        # Compute coordinate differences between successive slices
        origin = None
        origin_diff = np.array((0, 0, 0), dtype=float)
        origin_seen = []
        n_diffs = 0
        for file_origin in self._iter_origins():
            if any(np.allclose(file_origin, o) for o in origin_seen):
                continue  # ignore duplicates (seen in 4D images)

            if origin is not None:
                diff = file_origin - origin
                origin_diff = origin_diff + diff
                n_diffs += 1
            origin = file_origin
            origin_seen.append(file_origin)

        if n_diffs == 0:
            # One slice or volume only, read spacing from header or default to 1 mm
            z_i = float(getattr(self.ref_header, "SpacingBetweenSlices", 1.0))
            z_order = 0
        else:
            # Multiple slices, average spacing between slices
            avg_origin_diff = origin_diff / n_diffs
            z_i = np.linalg.norm(avg_origin_diff)

            # Use orientation of the coordinate system to determine in which
            # direction the origins of the individual slices should move.
            # Use the dot product to find the angle between that direction
            # and the spacing vector - this tells us whether the order of
            # the slices is correct or should be reversed
            z_direction = self.direction @ np.ones(self.dimensions)
            if self.dimensions > 3:
                avg_origin_diff = np.pad(
                    avg_origin_diff,
                    pad_width=(0, self.dimensions - avg_origin_diff.size),
                    mode="constant",
                )
            z_order = np.sign(np.dot(avg_origin_diff, z_direction))

        try:
            pixel_spacing = _find_dicom_tag(self.ref_header, "PixelSpacing")
        except DicomTagNotFoundError:
            pixel_spacing = (1.0, 1.0)

        x_i, y_i = (float(s) for s in pixel_spacing)
        spacing = (x_i, y_i, z_i)

        # Slices might require reordering, so compute the actual origin
        if origin is None:
            origin = (0.0, 0.0, 0.0)
        else:
            origin = tuple(origin_seen[0] if z_order >= 0 else origin)

        # Add dummy values for 4D images
        if self.dimensions == 4:
            spacing += (1.0,)
            origin += (0.0,)

        return origin, spacing, z_order

    def _pixel_values_need_scaling(self) -> bool:
        apply_slope = any(
            not isclose(float(getattr(h["data"], "RescaleSlope", 1.0)), 1.0)
            for h in self.headers
        )
        apply_intercept = any(
            not isclose(
                float(getattr(h["data"], "RescaleIntercept", 0.0)), 0.0
            )
            for h in self.headers
        )
        return apply_slope or apply_intercept

    def _read_pixel_values(self, filename: Path, rescale: bool) -> np.ndarray:
        ds = pydicom.dcmread(str(filename))

        # Read rescaling parameters already now so that we can delete
        # the DICOM dataset instance as soon as possible
        slope = float(getattr(ds, "RescaleSlope", 1))
        intercept = float(getattr(ds, "RescaleIntercept", 0))

        # If the data type is still unknown, use data type of this slice
        if self._pixel_value_dtype is None:
            if rescale:
                self._pixel_value_dtype = np.float32
            else:
                self._pixel_value_dtype = ds.pixel_array.dtype

        # Get pixel array and cast to desired data type if needed
        pixel_array = ds.pixel_array.astype(
            dtype=self._pixel_value_dtype, copy=False
        )
        del ds

        if rescale:
            return slope * pixel_array + intercept
        else:
            return pixel_array

    def _shape(self, samples_per_pixel: int) -> tuple[int, ...]:
        pixel_dims = [
            self.n_slices,
            self.ref_header.Rows,
            self.ref_header.Columns,
        ]
        if self.dimensions == 4:
            pixel_dims.insert(0, self.n_time)
        if samples_per_pixel > 1:
            pixel_dims.append(samples_per_pixel)

        return tuple(int(s) for s in pixel_dims)

    def _create_itk_from_dcm(self, z_order: int) -> SimpleITK.Image:
        samples_per_pixel = int(getattr(self.ref_header, "SamplesPerPixel", 1))
        pixel_dims = self._shape(samples_per_pixel)
        is_rgb = samples_per_pixel > 1
        apply_scaling = self._pixel_values_need_scaling()
        dcm_array: np.ndarray = None

        for index, partial in enumerate(self.headers):
            pixel_array = self._read_pixel_values(
                filename=partial["file"], rescale=apply_scaling
            )

            if (
                len(pixel_array.shape) == self.dimensions
                or pixel_array.shape == pixel_dims
            ):
                slicing = [slice(None)] * self.dimensions
                slicing[-3] = slice(None, None, 1 if z_order >= 0 else -1)
                dcm_array = pixel_array[tuple(slicing)]
                break

            if dcm_array is None:
                dcm_array = np.zeros(pixel_dims, dtype=pixel_array.dtype)

            # Determine slice position in array based on slice order
            z_index = index if z_order >= 0 else len(self.headers) - index - 1

            # Add slice to volume
            if self.dimensions == 4:
                z = (
                    z_index % self.n_slices
                    if self.n_slices_per_file == 1
                    else slice(None)
                )
                dcm_array[index // self.n_slices, z, :, :] = pixel_array
            else:
                dcm_array[z_index % self.n_slices, :, :] = pixel_array

        # When images have a photometric interpretation that requires low
        # values to be displayed in white, we invert the image values
        photometric_interpretation = getattr(
            self.ref_header, "PhotometricInterpretation", None
        )
        if not is_rgb and photometric_interpretation == "MONOCHROME1":
            self._pixel_value_inverter = PixelValueInverter(dcm_array)
            dcm_array = self._pixel_value_inverter.invert(dcm_array)

        return SimpleITK.GetImageFromArray(dcm_array, isVector=is_rgb)

    def _add_optional_metadata(self, img: SimpleITK.Image):
        for f in OPTIONAL_METADATA_FIELDS:
            value = getattr(self.ref_header, f, "")
            str_value = str(value)

            # Skip empty entries, ITK will not write them anyway
            if str_value == "":
                continue

            validate_metadata_value(key=f, value=value)

            # Invert value of window level
            if f == "WindowCenter" and self._pixel_value_inverter is not None:
                centers = self._pixel_value_inverter.invert(value)
                str_value = str(
                    list(centers) if centers.size > 1 else centers.item()
                )

            img.SetMetaData(f, str_value)

    def _add_temporal_metadata(self, img: SimpleITK.Image, z_order: int):
        content_times = []
        exposures = []

        for index, partial in enumerate(self.headers):
            if self.n_slices_per_file > 1 or index % self.n_slices == 0:
                content_times.append(str(partial["data"].ContentTime))
                exposures.append(str(partial["data"].Exposure))

        if z_order < 0:
            content_times = content_times[::-1]
            exposures = exposures[::-1]

        img.SetMetaData("ContentTimes", " ".join(content_times))
        img.SetMetaData("Exposures", " ".join(exposures))

    def read(self) -> SimpleITKImage:
        # Sort slices by instance number, which might be result in an order
        # in which slices are ordered from low to high coordinates according
        # to the default DICOM coordinate system - or the other way around.
        # First sort the slices, then find out which of these two options it
        # is (z_order will be 1 or -1, or 0 if there is a single slice)
        self._sort_slices_by_instance_number()
        origin, spacing, z_order = self._determine_slice_order()

        # Create ITK image from DICOM
        img = self._create_itk_from_dcm(z_order=z_order)
        img.SetDirection(tuple(self.direction.flatten()))
        img.SetSpacing(spacing)
        img.SetOrigin(origin)

        # Add additional metadata
        self._add_optional_metadata(img)
        if self.dimensions == 4:
            self._add_temporal_metadata(img, z_order)

        return SimpleITKImage(
            image=img,
            name=self.name,
            consumed_files={d["file"] for d in self.headers},
            spacing_valid=True,
        )


def get_dicom_headers_by_study(
    files: set[Path], file_errors: DefaultDict[Path, list[str]]
):
    """
    Gets all headers from dicom files found in path.

    Parameters
    ----------
    files
        Paths images that were uploaded during an upload session.

    file_errors
        Dictionary in which reading errors are recorded per file

    Returns
    -------
    A dictionary of headers for all dicom image files found within path,
    grouped by study id.
    """
    study_key_type = tuple[str, ...]
    studies: dict[study_key_type, dict[str, Any]] = {}
    indices: dict[str, dict[study_key_type, int]] = {}

    for file in files:
        if not file.is_file():
            continue
        with file.open("rb") as f:
            try:
                # Read header only, skip reading the pixel data for now
                ds = pydicom.dcmread(f, stop_before_pixels=True)

                # Group by series instance uid or by stack ID (for 4D images)
                # Additionally also group by SOP class UID to skip over extra
                # raw data (dose reports for example) that are sometimes stored
                # under the same series instance UID.
                key: study_key_type = (
                    ds.StudyInstanceUID,
                    getattr(ds, "StackID", ds.SeriesInstanceUID),
                    ds.SOPClassUID,
                )

                studies[key] = studies.get(key, {})
                indices[ds.StudyInstanceUID] = indices.get(
                    ds.StudyInstanceUID, {}
                )

                try:
                    index = indices[ds.StudyInstanceUID][key]
                except KeyError:
                    index = len(indices[ds.StudyInstanceUID])
                    indices[ds.StudyInstanceUID][key] = index

                headers = studies[key].get("headers", [])
                headers.append({"file": file, "data": ds})
                studies[key]["headers"] = headers

                # Since we might need to combine multiple images with different
                # series instance UID (in 4D images), we cannot use the series
                # as the unique file name - instead, we use the study instance
                # uid and a counter (index) per study
                studies[key]["name"] = f"{ds.StudyInstanceUID}-{index}"

            except Exception as e:
                file_errors[file].append(format_error(str(e)))

    return studies


def _find_valid_dicom_files(
    files: set[Path], file_errors: DefaultDict[Path, list[str]]
) -> list[DicomDataset]:
    """
    Gets the headers for all dicom files on path and validates them.

    Parameters
    ----------
    files
        Paths images that were uploaded during an upload session.

    file_errors
        Dictionary in which reading errors are recorded per file

    Returns
    -------
    A list of `DicomDataset` instances

    Any study with an inconsistent amount of slices per time point is discarded.
    """
    studies = get_dicom_headers_by_study(files=files, file_errors=file_errors)
    result = []
    for key in studies:
        headers = studies[key]["headers"]
        set_name = studies[key]["name"]
        if not headers:
            continue

        n_files = len(headers)
        n_time = len(
            {
                int(header["data"].TemporalPositionIndex)
                for header in headers
                if "TemporalPositionIndex" in header["data"]
            }
        )
        sop_class_uids = [header["data"].SOPClassUID for header in headers]

        arbitrary_header = headers[0]["data"]
        try:
            n_slices_per_file = len(
                arbitrary_header.PerFrameFunctionalGroupsSequence
            )
        except AttributeError:
            n_slices_per_file = int(
                getattr(arbitrary_header, "NumberOfFrames", 1)
            )
        n_slices = n_files * n_slices_per_file

        if "1.2.840.10008.5.1.4.1.1.77.1.6" in sop_class_uids:
            for d in headers:
                file_errors[d["file"]].append(
                    format_error("WSI-DICOM not supported by DICOM builder")
                )
        elif n_time < 2:
            # Not a 4d dicom file
            result.append(
                DicomDataset(
                    name=set_name,
                    headers=headers,
                    n_time=None,
                    n_slices=n_slices,
                    n_slices_per_file=n_slices_per_file,
                )
            )
        elif len(headers) % n_time > 0:
            # Invalid 4d dicom file
            for d in headers:
                file_errors[d["file"]].append(
                    format_error("Number of slices per time point differs")
                )
        else:
            # Valid 4d dicom file
            result.append(
                DicomDataset(
                    name=set_name,
                    headers=headers,
                    n_time=n_time,
                    n_slices=n_slices // n_time,
                    n_slices_per_file=n_slices_per_file,
                )
            )

    del studies
    return result


def image_builder_dicom(*, files: set[Path]) -> Iterator[SimpleITKImage]:
    """
    Constructs image objects by inspecting files in a directory.

    Parameters
    ----------
    files
        Paths to images that were uploaded during an upload session.

    Returns
    -------
    An `ImageBuilder` object consisting of:
     - a list of filenames for all files consumed by the image builder
     - a list of detected images
     - a list files associated with the detected images
     - path->error message map describing what is wrong with a given file
    """
    file_errors: DefaultDict[Path, list[str]] = defaultdict(list)

    studies = _find_valid_dicom_files(files=files, file_errors=file_errors)
    for dicom_ds in studies:
        try:
            yield dicom_ds.read()
        except Exception as e:
            for d in dicom_ds.headers:
                file_errors[d["file"]].append(format_error(str(e)))

    if file_errors:
        raise UnconsumedFilesException(file_errors=file_errors)



from collections import defaultdict
from collections.abc import Iterable, Iterator
from pathlib import Path
from typing import Any, DefaultDict, Union

import numpy as np
import numpy.typing as npt
import SimpleITK
from construct.core import Float64l, Int8ul, PaddedString, StreamError, Struct
from pydantic import BaseModel

from panimg.contrib.oct_converter.image_types import (
    FundusImageWithMetaData,
    OCTVolumeWithMetaData,
)
from panimg.contrib.oct_converter.readers import E2E, FDA, FDS
from panimg.exceptions import UnconsumedFilesException, ValidationError
from panimg.models import EyeChoice, SimpleITKImage


def format_error(message: str) -> str:
    return f"OCT image builder: {message}"


LATERALITY_TO_EYE_CHOICE = defaultdict(
    lambda: EyeChoice.UNKNOWN,
    {"L": EyeChoice.OCULUS_SINISTER, "R": EyeChoice.OCULUS_DEXTER},
)

OCT_CONVERTER_TYPE = Union[FDS, FDA, E2E]


class OctDimensions(BaseModel):
    extent_x_mm: float
    resolution_y_mm: float
    extent_z_mm: float


def _create_itk_images(
    *,
    file: tuple[Path, str],
    oct_volumes: Iterable[OCTVolumeWithMetaData],
    fundus_images: Iterable[FundusImageWithMetaData],
    oct_slice_size: OctDimensions,
) -> Iterator[SimpleITKImage]:
    filepath = file[0]
    filetype = file[1]

    for volume in oct_volumes:
        eye_choice = LATERALITY_TO_EYE_CHOICE[volume.laterality]

        yield _create_itk_oct_volume(
            file=filepath,
            volume=volume.volume,
            oct_slice_size=oct_slice_size,
            eye_choice=eye_choice,
        )
    for image in fundus_images:
        eye_choice = LATERALITY_TO_EYE_CHOICE[image.laterality]

        if filetype != "e2e":
            img_array = image.image.astype(np.uint8)
            img_array = img_array[:, :, ::-1]
            is_vector = True
        else:
            img_array = image.image
            is_vector = False

        yield _create_itk_fundus_image(
            file=filepath,
            image=img_array,
            eye_choice=eye_choice,
            is_vector=is_vector,
        )


def _create_itk_oct_volume(
    *,
    file: Path,
    volume: list[npt.NDArray[Any]],
    oct_slice_size: OctDimensions,
    eye_choice: EyeChoice,
) -> SimpleITKImage:
    img_array = np.array(volume)
    img = SimpleITK.GetImageFromArray(img_array, isVector=False)
    [st, _, sl] = img_array.shape
    img.SetSpacing(
        [
            oct_slice_size.extent_x_mm / sl,
            oct_slice_size.resolution_y_mm,
            oct_slice_size.extent_z_mm / st,
        ]
    )
    return SimpleITKImage(
        image=img,
        name=file.name,
        consumed_files={file},
        spacing_valid=True,
        eye_choice=eye_choice,
    )


def _create_itk_fundus_image(
    *,
    file: Path,
    image: npt.NDArray[Any],
    eye_choice: EyeChoice,
    is_vector: bool,
) -> SimpleITKImage:
    img = SimpleITK.GetImageFromArray(image, isVector=is_vector)

    return SimpleITKImage(
        image=img,
        name=file.stem + "_fundus" + file.suffix,
        consumed_files={file},
        spacing_valid=False,
        eye_choice=eye_choice,
    )


def _extract_slice_size(*, img: Union[FDS, FDA]) -> OctDimensions:
    slice_size_meta_data = Struct(
        "unknown" / PaddedString(12, "utf16"),
        "xmm" / Float64l,
        "zmm" / Float64l,
        "y" / Float64l,
        "unknown6" / Float64l,
        "unknown7" / Float64l,
        "unknown8" / Int8ul,
    )
    with open(img.filepath, "rb") as f:
        chunk_location, chunk_size = img.chunk_dict[b"@PARAM_SCAN_04"]
        f.seek(chunk_location)
        raw = f.read(chunk_size)
        spacing = slice_size_meta_data.parse(raw)
        return OctDimensions(
            extent_x_mm=spacing.xmm,
            resolution_y_mm=spacing.y / 1000.0,
            extent_z_mm=spacing.zmm,
        )


def _get_image(*, file: Path) -> tuple[
    Iterable[OCTVolumeWithMetaData],
    Iterable[FundusImageWithMetaData],
    OctDimensions,
    tuple[Path, str],
]:
    with open(file, "rb") as f:
        header = f.read(7)

        if b"FDS" in header:
            fds_img = FDS(file)
            file_info = file, "fds"
            oct_slice_size = _extract_slice_size(img=fds_img)
            return (
                [fds_img.read_oct_volume()],
                [fds_img.read_fundus_image()],
                oct_slice_size,
                file_info,
            )
        elif b"FDA" in header:
            fda_img = FDA(file)
            file_info = file, "fda"
            oct_slice_size = _extract_slice_size(img=fda_img)
            return (
                [fda_img.read_oct_volume()],
                [fda_img.read_fundus_image()],
                oct_slice_size,
                file_info,
            )
        elif b"CMD" in header:
            e2e_img = E2E(file)
            file_info = file, "e2e"
            # We were unable to retrieve slice size information from the e2e files.
            # The following default values are taken from:
            # https://bitbucket.org/uocte/uocte/wiki/Heidelberg%20File%20Format
            oct_slice_size = OctDimensions(
                extent_x_mm=6, resolution_y_mm=0.0039, extent_z_mm=4.5
            )
            # Note that the return types from oct_converter are different
            # for e2e files
            return (
                e2e_img.read_oct_volume(),
                e2e_img.read_fundus_image(),
                oct_slice_size,
                file_info,
            )
        else:
            raise ValueError


def image_builder_oct(*, files: set[Path]) -> Iterator[SimpleITKImage]:
    """
    Constructs OCT image objects by inspecting files in a directory by using
    OCT-converter PyPI library.

    Parameters
    ----------
    path
        Path to a directory that contains all images that were uploaded during
        an upload session.

    Returns
    -------
    An `ImageBuilder` object consisting of:
     - a list of filenames for all files consumed by the image builder
     - a list of detected images
     - a list files associated with the detected images
     - path->error message map describing what is wrong with a given file
    """
    file_errors: DefaultDict[Path, list[str]] = defaultdict(list)

    for file in files:
        try:
            oct_volumes, fundus_images, oct_slice_size, file_out = _get_image(
                file=file
            )

            yield from _create_itk_images(
                file=file_out,
                oct_volumes=oct_volumes,
                # Skip fundus_images until we decide how to handle these
                fundus_images=[],  # fundus_images,
                oct_slice_size=oct_slice_size,
            )

        except (OSError, ValidationError, StreamError, ValueError, IndexError):
            file_errors[file].append(
                format_error(
                    "Not a valid OCT file (supported formats: .fds,.fda,.e2e)"
                )
            )

    if file_errors:
        raise UnconsumedFilesException(file_errors=file_errors)



import logging
from collections import defaultdict
from collections.abc import Iterable
from pathlib import Path
from typing import DefaultDict, Optional

from panimg.exceptions import UnconsumedFilesException
from panimg.image_builders import DEFAULT_IMAGE_BUILDERS
from panimg.models import PanImg, PanImgFile, PanImgResult, PostProcessorResult
from panimg.post_processors import DEFAULT_POST_PROCESSORS
from panimg.types import ImageBuilder, PostProcessor

logger = logging.getLogger(__name__)


def convert(
    *,
    input_directory: Path,
    output_directory: Path,
    builders: Optional[Iterable[ImageBuilder]] = None,
    post_processors: Optional[Iterable[PostProcessor]] = None,
    recurse_subdirectories: bool = True,
) -> PanImgResult:
    new_images: set[PanImg] = set()
    new_image_files: set[PanImgFile] = set()
    consumed_files: set[Path] = set()
    file_errors: DefaultDict[Path, list[str]] = defaultdict(list)

    builders = builders if builders is not None else DEFAULT_IMAGE_BUILDERS

    builder_names = ", ".join(str(b) for b in builders)
    logger.info(f"Using builders {builder_names}")

    _convert_directory(
        input_directory=input_directory,
        output_directory=output_directory,
        builders=builders,
        consumed_files=consumed_files,
        new_images=new_images,
        new_image_files=new_image_files,
        file_errors=file_errors,
        recurse_subdirectories=recurse_subdirectories,
    )

    result = post_process(
        image_files=new_image_files,
        post_processors=(
            post_processors
            if post_processors is not None
            else DEFAULT_POST_PROCESSORS
        ),
    )
    new_image_files |= result.new_image_files

    return PanImgResult(
        new_images=new_images,
        new_image_files=new_image_files,
        consumed_files=consumed_files,
        file_errors=file_errors,
    )


def _convert_directory(
    *,
    input_directory: Path,
    output_directory: Path,
    builders: Iterable[ImageBuilder],
    consumed_files: set[Path],
    new_images: set[PanImg],
    new_image_files: set[PanImgFile],
    file_errors: DefaultDict[Path, list[str]],
    recurse_subdirectories: bool = True,
):
    input_directory = Path(input_directory).resolve()
    output_directory = Path(output_directory).resolve()

    output_directory.mkdir(exist_ok=True)

    files = set()
    for o in Path(input_directory).iterdir():
        if o.is_dir() and recurse_subdirectories:
            _convert_directory(
                input_directory=input_directory
                / o.relative_to(input_directory),
                output_directory=output_directory
                / o.relative_to(input_directory),
                builders=builders,
                consumed_files=consumed_files,
                new_images=new_images,
                new_image_files=new_image_files,
                file_errors=file_errors,
                recurse_subdirectories=recurse_subdirectories,
            )
        elif o.is_file():
            files.add(o)

    logger.info(f"Converting directory {input_directory}")

    for builder in builders:
        builder_files = files - consumed_files

        if len(builder_files) == 0:
            return

        logger.debug(f"Processing {len(builder_files)} file(s) with {builder}")

        builder_result = _build_files(
            builder=builder,
            files=builder_files,
            output_directory=output_directory,
        )

        new_images |= builder_result.new_images
        new_image_files |= builder_result.new_image_files
        consumed_files |= builder_result.consumed_files

        if builder_result.consumed_files:
            logger.info(
                f"{builder} created {len(builder_result.new_images)} "
                f"new images(s) from {len(builder_result.consumed_files)} file(s)"
            )
        else:
            logger.debug(f"No files consumed by {builder}")

        for filepath, errors in builder_result.file_errors.items():
            file_errors[filepath].extend(errors)


def _build_files(
    *, builder: ImageBuilder, files: set[Path], output_directory: Path
) -> PanImgResult:
    new_images = set()
    new_image_files: set[PanImgFile] = set()
    consumed_files: set[Path] = set()
    file_errors: dict[Path, list[str]] = {}

    try:
        for result in builder(files=files):
            n_image, n_image_files = result.save(
                output_directory=output_directory
            )

            new_images.add(n_image)
            new_image_files |= n_image_files
            consumed_files |= result.consumed_files

    except UnconsumedFilesException as e:
        file_errors = e.file_errors

    return PanImgResult(
        new_images=new_images,
        new_image_files=new_image_files,
        consumed_files=consumed_files,
        file_errors=file_errors,
    )


def post_process(
    *, image_files: set[PanImgFile], post_processors: Iterable[PostProcessor]
) -> PostProcessorResult:
    """
    Run a set of post processors on a set of image files

    Post processors add new files and directories to existing images,
    such as DZI creation for TIFF images, or thumbnail generation.
    They do not produce new image entities.
    """
    new_image_files: set[PanImgFile] = set()

    logger.info(f"Post processing {len(image_files)} image(s)")

    existing_ids = {f.image_id for f in image_files}

    for processor in post_processors:
        result = processor(image_files=image_files)

        # Filter out any new image ids
        filtered_files = {
            f for f in result.new_image_files if f.image_id in existing_ids
        }

        excluded_files = result.new_image_files - filtered_files

        if excluded_files:
            logger.warning(f"Ignoring: {excluded_files}")

        new_image_files |= filtered_files

    return PostProcessorResult(new_image_files=new_image_files)



from pathlib import Path

import numpy as np
from construct import Int32un, PaddedString, Struct
from pylibjpeg import decode

from panimg.contrib.oct_converter.image_types import (
    FundusImageWithMetaData,
    OCTVolumeWithMetaData,
)


class FDA:
    """Class for extracting data from Topcon's .fda file format.

    Notes:
        Mostly based on description of .fda file format here:
        https://bitbucket.org/uocte/uocte/wiki/Topcon%20File%20Format

    Attributes:
        filepath (str): Path to .img file for reading.
        header (obj:Struct): Defines structure of volume's header.
        oct_header (obj:Struct): Defines structure of OCT header.
        fundus_header (obj:Struct): Defines structure of fundus header.
        chunk_dict (dict): Name of data chunks present in the file, and their start locations.
    """

    def __init__(self, filepath):
        self.filepath = Path(filepath)
        if not self.filepath.exists():
            raise FileNotFoundError(self.filepath)
        self.header = Struct(
            "FOCT" / PaddedString(4, "ascii"),
            "FDA" / PaddedString(3, "ascii"),
            "version_info_1" / Int32un,
            "version_info_2" / Int32un,
        )
        self.oct_header = Struct(
            "type" / PaddedString(1, "ascii"),
            "unknown1" / Int32un,
            "unknown2" / Int32un,
            "width" / Int32un,
            "height" / Int32un,
            "number_slices" / Int32un,
            "unknown3" / Int32un,
        )

        self.oct_header_2 = Struct(
            "unknown" / PaddedString(1, "ascii"),
            "width" / Int32un,
            "height" / Int32un,
            "bits_per_pixel" / Int32un,
            "number_slices" / Int32un,
            "unknown" / PaddedString(1, "ascii"),
            "size" / Int32un,
        )

        self.fundus_header = Struct(
            "width" / Int32un,
            "height" / Int32un,
            "bits_per_pixel" / Int32un,
            "number_slices" / Int32un,
            "unknown" / PaddedString(4, "ascii"),
            "size" / Int32un,
            # 'img' / Int8un,
        )

        self.chunk_dict = self.get_list_of_file_chunks()

    def get_list_of_file_chunks(self):
        """Find all data chunks present in the file.

        Returns:
            dict
        """
        chunk_dict = {}
        with open(self.filepath, "rb") as f:
            # skip header
            raw = f.read(15)
            header = self.header.parse(raw)

            eof = False
            while not eof:
                chunk_name_size = np.frombuffer(f.read(1), dtype=np.uint8)[0]
                if chunk_name_size == 0:
                    eof = True
                else:
                    chunk_name = f.read(chunk_name_size)
                    chunk_size = np.frombuffer(f.read(4), dtype=np.uint32)[0]
                    chunk_location = f.tell()
                    f.seek(chunk_size, 1)
                    chunk_dict[chunk_name] = [chunk_location, chunk_size]
        print(f"File {self.filepath} contains the following chunks:")
        for key in chunk_dict.keys():
            print(key)
        return chunk_dict

    def read_oct_volume(self):
        """Reads OCT data.

        Returns:
            obj:OCTVolumeWithMetaData
        """

        if b"@IMG_JPEG" not in self.chunk_dict:
            raise ValueError(
                "Could not find OCT header @IMG_JPEG in chunk list"
            )
        with open(self.filepath, "rb") as f:
            chunk_location, chunk_size = self.chunk_dict[b"@IMG_JPEG"]
            f.seek(chunk_location)  # Set the chunk’s current position.
            raw = f.read(25)
            oct_header = self.oct_header.parse(raw)
            volume = np.zeros(
                (oct_header.height, oct_header.width, oct_header.number_slices)
            )
            for i in range(oct_header.number_slices):
                size = np.frombuffer(f.read(4), dtype=np.int32)[0]
                raw_slice = f.read(size)
                slice = decode(raw_slice)
                volume[:, :, i] = slice
        oct_volume = OCTVolumeWithMetaData(
            [volume[:, :, i] for i in range(volume.shape[2])]
        )
        return oct_volume

    def read_fundus_image(self):
        """Reads fundus image.

        Returns:
            obj:FundusImageWithMetaData
        """
        if b"@IMG_FUNDUS" not in self.chunk_dict:
            raise ValueError(
                "Could not find fundus header @IMG_FUNDUS in chunk list"
            )
        with open(self.filepath, "rb") as f:
            chunk_location, chunk_size = self.chunk_dict[b"@IMG_FUNDUS"]
            f.seek(chunk_location)  # Set the chunk’s current position.
            raw = f.read(24)  # skip 24 is important
            fundus_header = self.fundus_header.parse(raw)
            number_pixels = fundus_header.width * fundus_header.height * 3
            raw_image = f.read(fundus_header.size)
            image = decode(raw_image)
        fundus_image = FundusImageWithMetaData(image)
        return fundus_image



import datetime
import logging
import re
import shutil
from enum import Enum
from functools import cached_property
from pathlib import Path
from typing import Any, NamedTuple, Optional
from uuid import UUID, uuid4

import numpy as np
import SimpleITK
from pydantic import BaseModel, ConfigDict, field_validator
from pydantic.dataclasses import dataclass
from SimpleITK import GetArrayViewFromImage, Image, WriteImage

from panimg.exceptions import ValidationError

logger = logging.getLogger(__name__)

MASK_TYPE_PIXEL_IDS = [
    SimpleITK.sitkInt8,
    SimpleITK.sitkUInt8,
]


class ColorSpace(str, Enum):
    GRAY = "GRAY"
    RGB = "RGB"
    RGBA = "RGBA"
    YCBCR = "YCBCR"


ITK_COLOR_SPACE_MAP = {
    1: ColorSpace.GRAY,
    3: ColorSpace.RGB,
    4: ColorSpace.RGBA,
}


class ImageType(str, Enum):
    MHD = "MHD"
    TIFF = "TIFF"
    DZI = "DZI"


class EyeChoice(str, Enum):
    OCULUS_DEXTER = "OD"
    OCULUS_SINISTER = "OS"
    UNKNOWN = "U"
    NOT_APPLICABLE = "NA"


class PatientSex(str, Enum):
    MALE = "M"
    FEMALE = "F"
    OTHER = "O"
    EMPTY = ""


DICOM_VR_TO_VALIDATION_REGEXP = {
    "AS": re.compile(r"^\d{3}[DWMY]$"),
    "CS": re.compile(r"^[A-Z\d _]{0,16}$"),
    "DA": re.compile(r"^\d{8}$"),
    "LO": re.compile(r"^[^\\]{0,64}$"),
    "PN": re.compile(r"^[^\\]{0,324}$"),
    "UI": re.compile(r"^[\d.]{0,64}$"),
}
DICOM_VR_TO_VALUE_CAST = {
    "DA": lambda v: datetime.date(int(v[:4]), int(v[4:6]), int(v[6:8]))
}

# NOTE: Only int8 or uint8 data types are checked for segments
# so the true maximum is 256
MAXIMUM_SEGMENTS_LENGTH = 64


class ExtraMetaData(NamedTuple):
    keyword: str  # DICOM tag keyword (eg. 'PatientID')
    vr: str  # DICOM Value Representation (eg. 'LO')
    field_name: str  # Name of field on PanImg model (eg. 'patient_id')
    default_value: Any  # Default value for field

    @property
    def match_pattern(self):
        return DICOM_VR_TO_VALIDATION_REGEXP[self.vr]

    @property
    def cast_func(self):
        def default_func(v):
            return None if v == "" else v

        return DICOM_VR_TO_VALUE_CAST.get(self.vr, default_func)

    def validate_value(self, value):
        if value is None or str(value) == "":
            return
        if not re.match(self.match_pattern, str(value)):
            raise ValidationError(
                f"Value {value!r} for field {self.keyword!r} does not match "
                f"pattern {self.match_pattern.pattern!r}"
            )
        try:
            self.cast_func(value)
        except ValueError as e:
            raise ValidationError from e


def validate_metadata_value(*, key, value):
    key_to_md = {md.keyword: md for md in EXTRA_METADATA}
    if key in key_to_md:
        key_to_md[key].validate_value(value)


EXTRA_METADATA = (
    ExtraMetaData("PatientID", "LO", "patient_id", ""),
    ExtraMetaData("PatientName", "PN", "patient_name", ""),
    ExtraMetaData("PatientBirthDate", "DA", "patient_birth_date", None),
    ExtraMetaData("PatientAge", "AS", "patient_age", ""),
    ExtraMetaData("PatientSex", "CS", "patient_sex", ""),
    ExtraMetaData("StudyDate", "DA", "study_date", None),
    ExtraMetaData("StudyInstanceUID", "UI", "study_instance_uid", ""),
    ExtraMetaData("SeriesInstanceUID", "UI", "series_instance_uid", ""),
    ExtraMetaData("StudyDescription", "LO", "study_description", ""),
    ExtraMetaData("SeriesDescription", "LO", "series_description", ""),
)


@dataclass(frozen=True)
class PanImg:
    pk: UUID
    name: str
    width: int
    height: int
    depth: Optional[int]
    voxel_width_mm: Optional[float]
    voxel_height_mm: Optional[float]
    voxel_depth_mm: Optional[float]
    timepoints: Optional[int]
    resolution_levels: Optional[int]
    window_center: Optional[float]
    window_width: Optional[float]
    color_space: ColorSpace
    eye_choice: EyeChoice
    patient_id: str = ""
    patient_name: str = ""
    patient_birth_date: Optional[datetime.date] = None
    patient_age: str = ""
    patient_sex: PatientSex = PatientSex.EMPTY
    study_date: Optional[datetime.date] = None
    study_instance_uid: str = ""
    series_instance_uid: str = ""
    study_description: str = ""
    series_description: str = ""
    segments: Optional[frozenset[int]] = None


@dataclass(frozen=True)
class PanImgFile:
    image_id: UUID
    image_type: ImageType
    file: Path
    directory: Optional[Path] = None


@dataclass
class PanImgResult:
    new_images: set[PanImg]
    new_image_files: set[PanImgFile]
    consumed_files: set[Path]
    file_errors: dict[Path, list[str]]


@dataclass
class PostProcessorResult:
    new_image_files: set[PanImgFile]


class SimpleITKImage(BaseModel):
    image: Image

    name: str
    consumed_files: set[Path]

    spacing_valid: bool
    eye_choice: EyeChoice = EyeChoice.NOT_APPLICABLE

    model_config = ConfigDict(arbitrary_types_allowed=True, frozen=True)

    @property
    def width(self) -> int:
        return int(self.image.GetWidth())

    @property
    def height(self) -> int:
        return int(self.image.GetHeight())

    @property
    def depth(self) -> Optional[int]:
        try:
            depth = int(self.image.GetDepth())
        except (RuntimeError, ValueError):
            return None

        return depth or None

    @staticmethod
    def _extract_first_float(value: str) -> float:
        if value.startswith("["):
            return float(value[1:-1].split(",")[0])
        else:
            return float(value)

    @property
    def window_center(self) -> Optional[float]:
        try:
            return self._extract_first_float(
                self.image.GetMetaData("WindowCenter")
            )
        except (RuntimeError, ValueError):
            return None

    @property
    def window_width(self) -> Optional[float]:
        try:
            return self._extract_first_float(
                self.image.GetMetaData("WindowWidth")
            )
        except (RuntimeError, ValueError):
            return None

    @property
    def timepoints(self) -> Optional[int]:
        if self.image.GetDimension() == 4 and self.segments is None:
            # Only 4D files that are non-segmentations have timepoints
            return int(self.image.GetSize()[3])
        else:
            return None

    @field_validator("image")
    @classmethod
    def check_color_space(cls, image: Image):  # noqa: B902, N805
        cs = image.GetNumberOfComponentsPerPixel()
        if cs not in ITK_COLOR_SPACE_MAP:
            raise ValueError(f"Unknown color space for MetaIO image: {cs}")
        return image

    @field_validator("image")
    @classmethod
    def add_value_range_meta_data(cls, image: Image):  # noqa: B902, N805
        smallest_tag = "SmallestImagePixelValue"
        largest_tag = "LargestImagePixelValue"

        if image.HasMetaDataKey(smallest_tag) and image.HasMetaDataKey(
            largest_tag
        ):
            return image

        # Use the numpy-array route to support a larger range of data-types
        # than ITK' MinimumMaximumImageFilter (e.g. 2D uint8)
        array = GetArrayViewFromImage(image)
        image.SetMetaData(smallest_tag, str(array.min()))
        image.SetMetaData(largest_tag, str(array.max()))

        return image

    @cached_property
    def segments(self) -> Optional[frozenset[int]]:
        if (
            self.image.GetNumberOfComponentsPerPixel() != 1
            or self.image.GetPixelIDValue() not in MASK_TYPE_PIXEL_IDS
        ):
            # Only single channel, 8 bit images should be checked.
            # Everything else is not a segmentation
            return None

        im_arr = GetArrayViewFromImage(self.image)

        if self.image.GetDimension() == 4:
            segments = set()
            n_volumes = self.image.GetSize()[3]

            for volume in range(n_volumes):
                # Calculate the segments for each volume for memory efficiency
                volume_segments = np.unique(im_arr[volume, :, :, :])
                segments.update({*volume_segments})

                if not segments.issubset({0, 1}):
                    # 4D Segmentations must only have values 0 and 1
                    # as the 4th dimension encodes the overlay type
                    return None

            # Use 1-indexing for each segmentation
            segments = {idx + 1 for idx in range(n_volumes)}
        else:
            segments = np.unique(im_arr)

        if len(segments) <= MAXIMUM_SEGMENTS_LENGTH:
            return frozenset(segments)
        else:
            return None

    @property
    def color_space(self) -> ColorSpace:
        return ITK_COLOR_SPACE_MAP[self.image.GetNumberOfComponentsPerPixel()]

    @property
    def voxel_width_mm(self) -> Optional[float]:
        if self.spacing_valid:
            return float(self.image.GetSpacing()[0])
        else:
            return None

    @property
    def voxel_height_mm(self) -> Optional[float]:
        if self.spacing_valid:
            return float(self.image.GetSpacing()[1])
        else:
            return None

    @property
    def voxel_depth_mm(self) -> Optional[float]:
        if self.spacing_valid:
            try:
                return float(self.image.GetSpacing()[2])
            except IndexError:
                return None
        else:
            return None

    def generate_extra_metadata(self) -> dict[str, Any]:
        extra_metadata = {
            md.field_name: md.default_value for md in EXTRA_METADATA
        }
        for md in EXTRA_METADATA:
            try:
                value = str(self.image.GetMetaData(md.keyword))
            except (RuntimeError, ValueError):
                pass
            else:
                try:
                    md.validate_value(value)
                    if str(value) != "":
                        extra_metadata[md.field_name] = md.cast_func(value)
                except ValidationError as e:
                    # Validation of metadata is already done in the builders so
                    # that it only fails and skips the images with corrupt
                    # metadata. This validation is done as an extra check.
                    logger.warning(
                        f"Value for metadata field {md.keyword!r} is stripped "
                        f"because it produced a ValidationError: {e!r}"
                    )
        return extra_metadata

    def save(self, output_directory: Path) -> tuple[PanImg, set[PanImgFile]]:
        pk = uuid4()

        work_dir = Path(output_directory) / self.name
        work_dir.mkdir()

        new_image = PanImg(
            pk=pk,
            name=self.name,
            width=self.width,
            height=self.height,
            depth=self.depth,
            window_center=self.window_center,
            window_width=self.window_width,
            timepoints=self.timepoints,
            resolution_levels=None,
            color_space=self.color_space,
            voxel_width_mm=self.voxel_width_mm,
            voxel_height_mm=self.voxel_height_mm,
            voxel_depth_mm=self.voxel_depth_mm,
            eye_choice=self.eye_choice,
            segments=self.segments,
            **self.generate_extra_metadata(),
        )

        WriteImage(
            image=self.image,
            fileName=str(work_dir.absolute() / f"{pk}.mha"),
            useCompression=True,
        )

        new_files = set()
        for file in work_dir.iterdir():
            new_file = PanImgFile(
                image_id=pk, image_type=ImageType.MHD, file=file
            )
            new_files.add(new_file)

        return new_image, new_files


class TIFFImage(BaseModel):
    file: Path

    name: str
    consumed_files: set[Path]

    width: int
    height: int
    voxel_width_mm: float
    voxel_height_mm: float
    resolution_levels: int
    color_space: ColorSpace
    eye_choice: EyeChoice = EyeChoice.NOT_APPLICABLE
    segments: Optional[frozenset[int]] = None

    model_config = ConfigDict(frozen=True)

    def save(self, output_directory: Path) -> tuple[PanImg, set[PanImgFile]]:
        pk = uuid4()

        output_file = output_directory / self.name / f"{pk}{self.file.suffix}"
        output_file.parent.mkdir()

        new_image = PanImg(
            pk=pk,
            name=self.name,
            width=self.width,
            height=self.height,
            depth=1,
            resolution_levels=self.resolution_levels,
            color_space=self.color_space,
            voxel_width_mm=self.voxel_width_mm,
            voxel_height_mm=self.voxel_height_mm,
            voxel_depth_mm=None,
            timepoints=None,
            window_center=None,
            window_width=None,
            eye_choice=self.eye_choice,
            segments=self.segments,
            **{md.field_name: md.default_value for md in EXTRA_METADATA},
        )

        shutil.copy(src=self.file, dst=output_file)

        new_files = {
            PanImgFile(
                image_id=pk,
                image_type=ImageType.TIFF,
                file=output_file.absolute(),
            )
        }

        return new_image, new_files
